﻿using System;
using System.Collections.Generic;
using NaGet.SubCommands.SubTask;
using NaGet.Tasks;

namespace NaGet.SubCommands
{
	/// <summary>
	/// NaGetタスク処理のパック
	/// </summary>
	public abstract class NaGetTaskSet2 : NaGetTaskSet
	{
		/// <summary>
		/// サブタスクのハンドラ
		/// </summary>
		public virtual event EventHandler<TaskEventArgs> SubTaskEventRaised;
		
		/// <summary>
		/// サブタスクのリスト
		/// </summary>
		protected IList<NaGetSubTask> subTasks;
		
		/// <summary>
		/// 現在実行中のサブタスクのインデックス
		/// </summary>
		private int currentSubTaskIndex = -1;
		
		/// <summary>
		/// 終了したか
		/// </summary>
		protected bool isDone = false;
		
		/// <summary>
		/// キャンセルされたか
		/// </summary>
		protected bool isCancelled = false;
		
		/// <summary>
		/// キャンセルが呼ばれたか
		/// </summary>
		protected bool cancelCalled = false;
		
		/// <summary>
		/// キャンセル呼び出し時のイベント。
		/// </summary>
		protected event Action<object> onCancelCalled;
		
		public NaGetTaskSet2()
		{
		}
		
		/// <summary>
		/// キャンセル処理を行う
		/// </summary>
		/// <returns>成功したか否か</returns>
		public override bool Cancel()
		{
			if (Cancelable) {
				if (! cancelCalled && ! isDone) {
					cancelCalled = true;
					if (onCancelCalled != null) {
						onCancelCalled(null);
					}
					
					if (currentSubTask != null && currentSubTask.Cancelable) {
						return currentSubTask.Cancel();
					} else {
						return true;
					}
				} else {
					return false;
				}
			} else {
				return false;
			}
		}
		
		public override int CurrentTaskSetIndex {
			get { return currentSubTaskIndex; }
		}
		
		public override bool Done {
			get { return isDone; }
		}
		
		public virtual bool Cancelled {
			get { return isCancelled; }
		}
		
		#region フラグ処理の便利メソッド
		
		/// <summary>
		/// 開始時に関するフラグの処理を行う
		/// </summary>
		protected virtual void NotifyStarted()
		{
			currentSubTaskIndex = 0;
			isDone = false;
		}
		
		/// <summary>
		/// キャンセル中断時に関するフラグの処理を行う
		/// </summary>
		protected virtual void NotifyCancelled()
		{
			isCancelled = true;
			isDone = true;
		}
		
		/// <summary>
		/// 成功終了時に関するフラグの処理を行う
		/// </summary>
		protected virtual void NotifyCompleted()
		{
			isDone = true;
		}
		
		/// <summary>
		/// サブタスク実行を次へ。
		/// </summary>
		protected virtual void NotifyGoToNextSubTask()
		{
			currentSubTaskIndex ++;
		}
		
		/// <summary>
		/// サブタスク実行をジャンプする。
		/// </summary>
		/// <param name="subTaskIndex">ジャンプ先のサブタスク番号</param>
		protected virtual void NotifyGoToSubTask(int subTaskIndex)
		{
			currentSubTaskIndex = subTaskIndex;
		}
		
		#endregion
		
		#region サブタスク実行時における便利メソッド
				
		protected virtual string currentSubTaskName {
			get { return taskSetNames[currentSubTaskIndex]; }
		}
		
		protected virtual NaGetSubTask currentSubTask {
			get { return subTasks[currentSubTaskIndex]; }
		}
		
		protected virtual bool hasMoreSubTask {
			get { return currentSubTaskIndex < taskSetNames.Count; }
		}
		
		#endregion
		
		#region サブタスク初期化・登録便利メソッド
		
		protected void initSubTask()
		{
			taskSetNames = new List<string>();
			subTasks = new List<NaGetSubTask>();
		}
		
		/// <summary>
		/// サブタスクを登録する
		/// </summary>
		/// <param name="name">名称</param>
		/// <param name="task">サブタスクオブジェクト</param>
		protected void registSubTask(string name, NaGetSubTask task)
		{
			taskSetNames.Add(name);
			subTasks.Add(task);
			
			// サブタスクのイベントをSubTaskEventRaised,TaskEventRaisedに転送
			task.TaskEventRaised += new EventHandler<TaskEventArgs>(subtaskEventTransformer);
		}
		
		/// <summary>
		/// サブタスクのイベントをSubTaskEventRaised,TaskEventRaisedに転送する関数
		/// </summary>
		/// <param name="sender">送信元のサブタスク</param>
		/// <param name="e">イベントオブジェクト</param>
		private void subtaskEventTransformer(object sender, TaskEventArgs e)
		{
			if (SubTaskEventRaised != null) {
				SubTaskEventRaised(sender, e);
			}
			if (e.ProgressPercent >= 0) {
				RaiseTaskSetEvent(TaskEventType.PING, null, GetProgressPercent(TaskEventType.PING, e.ProgressPercent));
			}
		}
		
		#endregion

	}
}
