﻿using System;
using System.IO;
using NaGet.Packages;
using NaGet.Packages.Install;
using NaGet.Net;
using NaGet.SubCommands;
using NaGet.Tasks;

namespace NaGet.SubCommands
{
	public class NaGetUpdate : NaGetTaskSet
	{
		private RepositoriesList repoList;
		
		private PackageListsManager pkgListMan;
		
		private bool downloadPackageLists = false;
		
		private bool packageListsDownloaded = true;
		
		
		/// <summary>
		/// ダウンロードに使うダウンローダオブジェクト
		/// </summary>
		public Downloader Downloader {
			get {
				if (downloader == null) {
					downloader = new Downloader();
					downloader.DownloadEventRaised += delegate(object sender, DownloadEventArgs e) {
						if (e.DownloadTaskType == DownloadEventType.DOWNLOADING && e.ProgressPercent > 0) {
							RaiseTaskSetEvent(TaskEventType.PING, string.Empty, GetProgressPercent(TaskEventType.PING, e.ProgressPercent));
						}
					};
				}
				return downloader;
			}
		}
		
		private Downloader downloader;
		
		private int currentTaskSetIndex = -1;
		
		private bool done = false;
		
		public override int CurrentTaskSetIndex {
			get { return currentTaskSetIndex; }
		}
		
		public override bool Cancelable {
			get { return ! packageListsDownloaded; }
		}
		
		public NaGetUpdate(PackageListsManager pkgListMan)
			: this(pkgListMan, true)
		{
		}
		
		public NaGetUpdate(PackageListsManager pkgMan, bool downloadPackageListsFlag)
		{
			pkgListMan = pkgMan;
			downloadPackageLists = downloadPackageListsFlag;
			
			taskSetNames = new System.Collections.Generic.List<string>();
			if (downloadPackageLists) {
				// repos.list.xmlがあるとき、そこからよみとる。
				repoList = NaGet.Utils.GetDeserializedObject<RepositoriesList>(NaGet.Env.RepositoriesListFile);
				
				foreach (RepositoryInfo repo in repoList.EnabledRepositories) {
					taskSetNames.Add(string.Format("リスト取得: {0}", repo.Url.Href));
				}
				taskSetNames.Add(string.Format("リスト更新: {0}", NaGet.Env.PackageListFile));
			}
			taskSetNames.Add(string.Format("リスト更新: {0}", NaGet.Env.ArchiveInstalledPackageListFile));
			taskSetNames.Add(string.Format("リスト更新: {0}", NaGet.Env.SystemInstalledPackageListFile));
		}

		public override void Run()
		{
			currentTaskSetIndex ++;
			RaiseTaskSetEvent(TaskEventType.STARTED, "リスト更新処理開始");
			
			try {
				// リストのダウンロード
				if (downloadPackageLists) {
					packageListsDownloaded = false;
					try {
						runDownloadPackageLists();
					} catch (TaskCanceledException) {
						RaiseTaskSetEvent(TaskEventType.WARNING, "リストのダウンロード処理がキャンセルされました");
						pkgListMan.LoadPackageLists();
					} catch (System.Net.WebException e) {
						RaiseTaskSetEvent(TaskEventType.WARNING, e.Message);
						if (System.Net.NetworkInformation.NetworkInterface.GetIsNetworkAvailable()) {
							RaiseTaskSetEvent(TaskEventType.WARNING, "ネットワークに接続されていません。");
						} else {
							RaiseTaskSetEvent(TaskEventType.WARNING, "ネットワークに接続できませんでした。ネットワークが切断されているか、ファイアウォールによって遮断された可能性があります。");
						}
						pkgListMan.LoadPackageLists();
					} finally {
						int numOfEnabledRepos = NaGet.Utils.IEnumerable2Array(repoList.EnabledRepositories).Length;
						currentTaskSetIndex = numOfEnabledRepos + 1;
					}
					packageListsDownloaded = true;
				}
				
				runLocalUpdate();
			} finally {
				done = true;
			}
			
			RaiseTaskSetEvent(TaskEventType.COMPLETED, "終了", 100);
		}
		
		private void runDownloadPackageLists()
		{
			PackageList<Package> avaiablePackageList = new PackageList<Package>();
			foreach(RepositoryInfo repo in repoList.EnabledRepositories) {
				RaiseTaskSetEvent(TaskEventType.STARTED_SUBTASK, taskSetNames[currentTaskSetIndex]);
				
				if (Uri.IsWellFormedUriString(repo.Url.Href, UriKind.Absolute)) {
					string tmpfileName = Path.GetTempFileName();
					try {
						Downloader.Download(repo.Url.Href, tmpfileName);
						if (repo.Type == RepositoryType.APPLISTATION_NATIVE_XML_1_0) {
							try {
								PackageList<Package> pkgList = NaGet.Utils.GetDeserializedObject<PackageList<Package>>(tmpfileName);
								pkgList.FixPackageListName(); // PackageListNameとの紐付けを行う
								
								// RepositoryReferenceの名前を読み込む // TODO RepositoryReferenceの名前を読み込む処理はここでいいのか?
								repo.Name = (string.IsNullOrEmpty(pkgList.Name))? repo.Name : pkgList.Name;
								
								avaiablePackageList.AddPackages(pkgList);
							} catch (InvalidOperationException) {
								RaiseTaskSetEvent(TaskEventType.ERROR, string.Format("レポジトリ'{0}'はAppliStation Native XML softlist形式ではありません。", repo.Name ?? repo.Url.Href));
							}
						} else {
							RaiseTaskSetEvent(TaskEventType.WARNING, string.Format("レポジトリ'{0}'の設定が不正です。", repo.Name ?? repo.Url.Href));
						}
					} finally {
						if (File.Exists(tmpfileName)) {
							File.Delete(tmpfileName);
						}
					}
				} else {
					RaiseTaskSetEvent(TaskEventType.WARNING, string.Format("レポジトリ'{0}'のURLが不正なため、ソフトリストは取得できませんでした。", repo.Name ?? repo.Url.Href));
				}
				
				RaiseTaskSetEvent(TaskEventType.COMPLETED_SUBTASK, taskSetNames[currentTaskSetIndex]);
				currentTaskSetIndex ++;
			}
			
			// TODO 暫定的にかならず常にrepositoryリストに書き込む。
			if ( true ) {
				NaGet.Utils.PutSerializeObject(NaGet.Env.RepositoriesListFile, repoList);
			}
			
			RaiseTaskSetEvent(TaskEventType.STARTED_SUBTASK, taskSetNames[currentTaskSetIndex]);
			pkgListMan.availablePkgList = avaiablePackageList; // Mediatorのリストを更新
			pkgListMan.SaveAvailablePackageList();
			RaiseTaskSetEvent(TaskEventType.COMPLETED_SUBTASK, taskSetNames[currentTaskSetIndex]);
			currentTaskSetIndex ++;
		}
		
		private void runLocalUpdate()
		{
			// インストールトリストの更新
			RaiseTaskSetEvent(TaskEventType.STARTED_SUBTASK, taskSetNames[currentTaskSetIndex]);
			pkgListMan.DetectInstalledPkgs();
			pkgListMan.SaveInstalledPackageList();
			RaiseTaskSetEvent(TaskEventType.COMPLETED_SUBTASK, TaskSetNames[currentTaskSetIndex]);
			currentTaskSetIndex++;
		
			// システムにインストールされているリストの更新
			RaiseTaskSetEvent(TaskEventType.STARTED_SUBTASK, TaskSetNames[currentTaskSetIndex]);
			pkgListMan.DetectSystemInstalledPkgs();
			pkgListMan.SaveSystemInstalledPackageList();
			RaiseTaskSetEvent(TaskEventType.COMPLETED_SUBTASK, TaskSetNames[currentTaskSetIndex]);
			currentTaskSetIndex++;
		}
		
		public override bool Cancel()
		{
			return Downloader.Cancel();
		}

		public override bool Done {
			get { return done; }
		}
	}
}
